#!/usr/bin/env python
#
# glmask_funcs.py - OpenGL 2.1 functions used by the GLMask class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLMask`
class to render :class:`.Image` overlays in an OpenGL 2.1 compatible manner.
"""


import numpy as np

import fsleyes.gl.shaders as shaders
from . import                glvolume_funcs


def init(self):
    """Calls the :func:`compileShaders` and :func:`updateShaderState`
    functions.
    """
    self.shader = None
    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Destroys the shader programs. """
    self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Loads the vertex/fragment shader source code, and creates a
    :class:`.GLSLShader` program.
    """

    if self.shader is not None:
        self.shader.destroy()

    vertSrc = shaders.getVertexShader(  'glvolume')
    fragSrc = shaders.getFragmentShader('glmask')

    self.shader = shaders.GLSLShader(vertSrc, fragSrc)


def updateShaderState(self):
    """Updates all shader program variables. """

    if not self.ready():
        return

    opts       = self.opts
    shader     = self.shader
    imageShape = np.array(self.image.shape[:3])
    vvx        = self.imageTexture.voxValXform

    shader.load()

    changed  = False
    changed |= shader.set('imageTexture',  0)
    changed |= shader.set('voxValXform',   vvx)
    changed |= shader.set('imageShape',    imageShape)
    changed |= shader.set('useSpline',     opts.interpolation == 'spline')
    changed |= shader.set('threshold',     self.getThreshold())
    changed |= shader.set('invert',        opts.invert)
    changed |= shader.set('colour',        self.getColour())

    shader.unload()

    return changed


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Draws a 2D slice at the given ``zpos``. Uses the
    :func:`.gl21.glvolume_funcs.draw2D` function.
    """
    self.shader.load()
    glvolume_funcs.draw2D(self, zpos, axes, xform, bbox)
    self.shader.unloadAtts()
    self.shader.unload()


def drawAll(self, axes, zposes, xforms):
    """Draws all specified slices. Uses the
    :func:`.gl21.glvolume_funcs.drawAll` function.
    """
    self.shader.load()
    glvolume_funcs.drawAll(self, axes, zposes, xforms)
    self.shader.unloadAtts()
    self.shader.unload()



#!/usr/bin/env python
#
# sampleline.py - The SampleLineAction
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SampleLineAction` class, which allows
the user to open a :class:`SampleLinePanel`. The ``SampleLinePanel`` is
a FSLeyes control which allows the user to draw a line on the canvases of
an :class:`.OrthoPanel`, and plot the data along that line from the currently
selected :class:`.Image` overlay.
"""

import                  os
import                  copy

import numpy         as np
import scipy.ndimage as ndimage
import                  wx

import fsl.data.image                             as fslimage
import fsl.utils.settings                         as fslsettings
import fsl.transform.affine                       as affine
import fsleyes_widgets.widgetlist                 as widgetlist
import fsleyes_widgets.utils.status               as status
import fsleyes_props                              as props
import fsleyes.strings                            as strings
import fsleyes.tooltips                           as tooltips
import fsleyes.icons                              as fslicons
import fsleyes.actions                            as actions
import fsleyes.actions.screenshot                 as screenshot
import fsleyes.views.orthopanel                   as orthopanel
import fsleyes.controls.controlpanel              as ctrlpanel
import fsleyes.plotting                           as plotting
import fsleyes.plotting.plotcanvas                as plotcanvas
import fsleyes.plugins.profiles.samplelineprofile as samplelineprofile


def sampleAlongLine(data, start, end, resolution, order):
    """Samples from ``data``, along a line between ``start`` and ``end``.

    :arg data:       3D array
    :arg start:      Start coordinate
    :arg end:        End coordinate
    :arg resolution: Number of points to sample
    :arg order:      Interpolation  (see ``scipy.ndimage.map_coordinates``)
    :returns:        Tuple containing:
                       - 1D Numpy array containing the sampled values
                       - ``(3, N)`` numpy array containing the coordinates for
                         each sample
    """

    start  = list(start)
    end    = list(end)
    shape  = data.shape
    coords = np.linspace(start, end, resolution).T

    # map_coordinates doesn't take
    # kindly to dims of length 1
    data = data.squeeze()
    if any(s == 1 for s in shape):
        drop      = [i for i, s in enumerate(shape) if s == 1]
        mapcoords = np.delete(coords, drop, axis=0)
    else:
        mapcoords = coords

    # multi-channel data?
    if len(data.dtype) > 1:
        channels = []
        for chan in data.dtype.fields.keys():
            channels.append(data[chan])
        data = channels
    else:
        data = [data]

    ys = []
    for arr in data:
        ys.append(ndimage.map_coordinates(arr,
                                          mapcoords,
                                          order=order,
                                          output=np.float64))

    # For multi channel data, we currently
    # just take the mean across all
    # channels, but this might change in
    # the future (if there is any need).
    if len(ys) > 1: y = np.mean(ys, axis=0)
    else:           y = ys[0]

    return y, coords


class SampleLineAction(actions.ToggleControlPanelAction):
    """The ``SampleLineAction`` simply shows/hides a :class:`SampleLinePanel`.
    """


    @staticmethod
    def supportedViews():
        """The ``SampleLineAction`` is restricted for use with
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    def __init__(self, overlayList, displayCtx, ortho):
        """Create a  ``SampleLineAction``. """
        super().__init__(overlayList, displayCtx, ortho, SampleLinePanel)

        self.__ortho = ortho
        self.__name  = '{}_{}'.format(type(self).__name__, id(self))

        displayCtx.addListener('selectedOverlay', self.__name,
                               self.__selectedOverlayChanged)


    def destroy(self):
        """Called when the :class:`.OrthoPanel` that owns this action is
        closed. Clears references, removes listeners, and calls the base
        class ``destroy`` method.
        """
        if self.destroyed:
            return

        self.__ortho = None
        self.displayCtx.removeListener('selectedOverlay', self.__name)
        super().destroy()


    def __selectedOverlayChanged(self, *a):
        """Called when the selected overlay changes. Enables/disables this
        action (and hence the bound Tools menu item) depending on whether the
        overlay is an image.
        """
        ovl = self.displayCtx.getSelectedOverlay()
        self.enabled = isinstance(ovl, fslimage.Image)


    def __run(self):
        """Open/close a :class:`SampleLinePanel`. """
        self.viewPanel.togglePanel(SampleLinePanel)


class SampleLineDataSeries(plotting.DataSeries):
    """The ``SampleLineDataSeries`` represents data that is sampled along
    a straight line through a 3D volume from an :class:`.Image` overlay.

    ``SampleLineDataSeries`` objects are created by the
    :class:`SampleLinePanel`.
    """


    interp = props.Choice((0, 1, 2, 3), default=0)
    """How to interpolate the sampled data when it is plotted. The value is
    used directly as the ``order`` parameter to the
    ``scipy.ndimage.map_coordinates`` function.
    """


    resolution = props.Int(minval=2, maxval=200, default=100, clamped=False)
    """The number of points (uniformly spaced) to sample along the line. """


    normalise = props.Choice(('none', 'y', 'x', 'xy'))
    """Whether to normalise all plotted data along the x, or y, or both, axes.
    """


    def __init__(self,
                 overlay,
                 overlayList,
                 displayCtx,
                 plotCanvas,
                 index,
                 start,
                 end):
        """Create a ``SampleLineDataSeries``.

        :arg overlay:     The :class:`.Image` overlay to sample from
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  A :class:`.DisplayContext` instance
        :arg plotCanvas:  The :class:`.PlotCanvas` that is plotting this
                          ``DataSeries``.
        :arg index:       Volume index, for images with more than 3
                          dimensions (see :meth:`.NiftiOpts.index`)
        :arg start:       Start of sampling line, in voxel coordinates
        :arg end:         End of sampling line, in voxel coordinates
        """

        plotting.DataSeries.__init__(
            self, overlay, overlayList, displayCtx, plotCanvas)

        self.__index  = index
        self.__start  = start
        self.__end    = end
        self.__coords = None

        self.addListener('resolution', self.name, self.__refreshData)
        self.addListener('interp',     self.name, self.__refreshData)
        self.addListener('normalise',  self.name, self.__refreshData)

        self.label = ('{}: [{:.2f} {:.2f} {:.2f}] -> '
                      '[{:.2f} {:.2f} {:.2f}]'.format(
                          overlay.name, *start, *end))


    @property
    def coords(self):
        """Return a ``(3, n)`` array containing the voxel coordinates of
        each sampled point for the most recently generated data, or ``None``
        if no data has been sampled yet.
        """
        return self.__coords


    def __refreshData(self, *a):
        """Called when :attr:`resolution`, :attr:`interp` or :attr:`normalise`
        change. Re-samples the data from the image.
        """

        overlay      = self.overlay
        opts         = self.displayCtx.getOpts(overlay)
        data         = overlay[self.__index]
        resolution   = self.resolution
        order        = self.interp
        normalisex   = 'x' in self.normalise
        normalisey   = 'y' in self.normalise
        start        = self.__start
        end          = self.__end

        y, coords = sampleAlongLine(data, start, end, resolution, order)

        if normalisey:
            y = (y - y.min()) / (y.max() - y.min())

        if normalisex:
            xmax = 1
        else:
            wstart = opts.transformCoords(start, 'voxel', 'world')
            wend   = opts.transformCoords(end,   'voxel', 'world')
            xmax   = affine.veclength(wstart - wend)[0]

        x = np.linspace(0, xmax, resolution)

        self.__coords = coords
        self.setData(x, y)


class SampleLinePanel(ctrlpanel.ControlPanel):
    """The ``SampleLinePanel`` is a FSLeyes control which can be used in
    conjunction with an :class:`.OrthoPanel` view. It allows the user to draw
    a line on an ``OrthoPanel`` canvas, and plot the voxel intensities, along
    that line, from the currently selected :class:`.Image` overlay.

    The :class:`.SampleLineProfile` class implements user interaction, and the
    :class:`.PlotCanvas` class is used for plotting.
    """


    # Used to synchronise between GUI widgets
    # and SampleLineDataSeries property values
    # The GUI widgets are bound to the most
    # recently added SampleLineDataSeries
    # instance
    interp     = copy.copy(SampleLineDataSeries.interp)
    resolution = copy.copy(SampleLineDataSeries.resolution)
    normalise  = copy.copy(SampleLineDataSeries.normalise)
    colour     = copy.copy(plotting.DataSeries.colour)
    lineWidth  = copy.copy(plotting.DataSeries.lineWidth)
    lineStyle  = copy.copy(plotting.DataSeries.lineStyle)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``SampleLinePanel`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    @staticmethod
    def ignoreControl():
        """Tells FSLeyes not to add the ``SampleLinePanel`` as an option to
        the Settings menu. Instead, the :class:`SampleLineAction` is added as
        an option to the Tools menu.
        """
        return True


    @staticmethod
    def profileCls():
        """Returns the :class:`.SampleLineProfile` class, which needs to be
        activated in conjunction with the ``SampleLinePanel``.
        """
        return samplelineprofile.SampleLineProfile


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'floatPane' : True,
                'floatOnly' : True}


    def __init__(self, parent, overlayList, displayCtx, ortho):
        """Create a ``SampleLinePanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, ortho)

        profile = ortho.currentProfile

        # plot which displays the sampled data
        canvas = plotcanvas.PlotCanvas(self, drawFunc=self.__draw)
        canvas.canvas.SetMinSize((-1, 150))

        self.__ortho   = ortho
        self.__profile = profile
        self.__canvas  = canvas
        self.__current = None

        # initial settings
        self.colour    = '#000050'
        self.lineWidth = 2
        canvas.legend  = False

        # Controls which allow the user to select
        # interpolation/resolution, line colour, etc
        widgets = widgetlist.WidgetList(self)
        legend = props.makeWidget(
            widgets, canvas, 'legend',
            labels=strings.properties['PlotCanvas.legend'])

        interp = props.makeWidget(
            widgets, self, 'interp', labels=strings.choices[self, 'interp'])
        resolution = props.makeWidget(
            widgets, self, 'resolution', slider=True, spin=True,
            showLimits=False)
        normalise = props.makeWidget(
            widgets, self, 'normalise',
            labels=strings.choices[self, 'normalise'])

        colour    = props.makeWidget(widgets, self, 'colour')
        lineWidth = props.makeWidget(widgets, self, 'lineWidth')
        lineStyle = props.makeWidget(
            widgets, self, 'lineStyle',
            labels=strings.choices['DataSeries.lineStyle'])

        widgets.AddWidget(legend,     strings.labels[self, 'legend'])
        widgets.AddWidget(interp,     strings.labels[self, 'interp'])
        widgets.AddWidget(resolution, strings.labels[self, 'resolution'])
        widgets.AddWidget(normalise,  strings.labels[self, 'normalise'])
        widgets.AddWidget(colour,     strings.labels[self, 'colour'])
        widgets.AddWidget(lineWidth,  strings.labels[self, 'lineWidth'])
        widgets.AddWidget(lineStyle,  strings.labels[self, 'lineStyle'])

        # Controls allowing the user to add/remove
        # sample lines from the plot to save the
        # data to a file, and to save a screenshot
        # of the plot
        ctrlSizer  = wx.BoxSizer(wx.HORIZONTAL)
        screenshot = actions.ActionButton(
            'screenshot',
            icon=fslicons.findImageFile('camera24'),
            tooltip=tooltips.actions['PlotPanel.screenshot'])
        export = actions.ActionButton(
            'export',
            icon=fslicons.findImageFile('exportDataSeries24'),
            tooltip=tooltips.actions['PlotPanel.exportDataSeries'])
        add = actions.ActionButton(
            'addDataSeries',
            icon=fslicons.findImageFile('add24'),
            tooltip=tooltips.actions[self, 'addDataSeries'])
        remove = actions.ActionButton(
            'removeDataSeries',
            icon=fslicons.findImageFile('remove24'),
            tooltip=tooltips.actions[self, 'removeDataSeries'])

        screenshot = props.buildGUI(self, self, screenshot)
        export     = props.buildGUI(self, self, export)
        add        = props.buildGUI(self, self, add)
        remove     = props.buildGUI(self, self, remove)

        ctrlSizer.Add(screenshot, flag=wx.EXPAND)
        ctrlSizer.Add(export,     flag=wx.EXPAND)
        ctrlSizer.Add(add,        flag=wx.EXPAND)
        ctrlSizer.Add(remove,     flag=wx.EXPAND)

        # Labels which show the start and end
        # coordinates of the sample line in voxel
        # ("v") and world ("w") coordinates, and
        # the line length in [units] (probably mm)
        vfromlbl = wx.StaticText(self)
        vtolbl   = wx.StaticText(self)
        wfromlbl = wx.StaticText(self)
        wtolbl   = wx.StaticText(self)
        vfromval = wx.StaticText(self)
        vtoval   = wx.StaticText(self)
        wfromval = wx.StaticText(self)
        wtoval   = wx.StaticText(self)
        lenlbl   = wx.StaticText(self)
        lenval   = wx.StaticText(self)

        self.__vfromval = vfromval
        self.__vtoval   = vtoval
        self.__wfromval = wfromval
        self.__wtoval   = wtoval
        self.__lenval   = lenval

        vfromlbl.SetLabel(strings.labels[self, 'voxelfrom'])
        vtolbl  .SetLabel(strings.labels[self, 'voxelto'])
        wfromlbl.SetLabel(strings.labels[self, 'worldfrom'])
        wtolbl  .SetLabel(strings.labels[self, 'worldto'])
        lenlbl  .SetLabel(strings.labels[self, 'length'])

        infoSizer = wx.FlexGridSizer(3, 4, 5, 5)
        infoSizer.AddGrowableCol(1)
        infoSizer.AddGrowableCol(3)
        infoSizer.Add(vfromlbl)
        infoSizer.Add(vfromval)
        infoSizer.Add(vtolbl)
        infoSizer.Add(vtoval)
        infoSizer.Add(wfromlbl)
        infoSizer.Add(wfromval)
        infoSizer.Add(wtolbl)
        infoSizer.Add(wtoval)
        infoSizer.Add(lenlbl)
        infoSizer.Add(lenval)
        infoSizer.Add((1, 1))
        infoSizer.Add((1, 1))

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        mainSizer.Add(ctrlSizer,     flag=wx.EXPAND)
        mainSizer.Add(widgets,       flag=wx.EXPAND)
        mainSizer.Add(infoSizer,     flag=wx.EXPAND)
        mainSizer.Add(canvas.canvas, flag=wx.EXPAND, proportion=1)

        self.SetSizer(mainSizer)
        self.Layout()

        profile.registerHandler('LeftMouseDown', self.name, self.__onMouseDown)
        profile.registerHandler('LeftMouseDrag', self.name, self.__onMouseDrag)
        profile.registerHandler('LeftMouseUp',   self.name, self.__onMouseUp)

        self.addListener('interp',     self.name, canvas.asyncDraw)
        self.addListener('resolution', self.name, canvas.asyncDraw)
        self.addListener('normalise',  self.name, canvas.asyncDraw)
        self.addListener('colour',     self.name, canvas.asyncDraw)
        self.addListener('lineWidth',  self.name, canvas.asyncDraw)
        self.addListener('lineStyle',  self.name, canvas.asyncDraw)


    def destroy(self):
        """Called when this ``SampleLinePanel`` is no longer needed. Clears
        references, and calls :meth:`.ControlPanel.destroy`.
        """
        super().destroy()
        self.__canvas.destroy()
        self.__ortho   = None
        self.__profile = None
        self.__canvas  = None
        self.__current = None


    @property
    def canvas(self):
        """Return a reference to the :class:`.PlotCanvas` that is used to
        plot sampled data from lines that the user has drawn.
        """
        return self.__canvas


    @actions.action
    def addDataSeries(self):
        """Holds/persists the most recently sampled line to the plot. """
        if self.__current is not None:
            self.__canvas.dataSeries.append(self.__current)
            self.__bindToDataSeries(self.__current, False)
            self.__current = None


    @actions.action
    def removeDataSeries(self):
        """Removes the most recently held/persisted line from the plot. """
        canvas = self.__canvas
        if len(canvas.dataSeries) > 0:
            canvas.dataSeries.pop()


    @actions.action
    def export(self):
        """Prompts the user to save the sampled data to a file. """

        # only one series can be saved - the
        # user is asked to select which one
        if self.__current is None: series = []
        else:                      series = [self.__current]

        series.extend(reversed(self.__canvas.dataSeries))

        if len(series) == 0:
            return

        # ask what series, and whether they
        # want coordinates of sample points
        parent = self.GetParent()
        dlg    = ExportSampledDataDialog(parent, series)
        if dlg.ShowModal() != wx.ID_OK:
            return

        series     = dlg.GetSeries()
        saveCoords = dlg.GetCoordinates()

        # ask where to save the file
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        msg     = strings.titles[self, 'savefile']
        dlg     = wx.FileDialog(parent,
                                message=msg,
                                defaultDir=fromDir,
                                defaultFile='sample.txt',
                                style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filename = dlg.GetPath()

        # prepare the data
        if saveCoords == 'none':
            data = series.getData()[1]
        else:
            image      = series.overlay
            opts       = self.displayCtx.getOpts(image)
            samples    = series.getData()[1]
            coords     = series.coords.T
            data       = np.zeros((len(samples), 4))
            data[:, 0] = samples

            if saveCoords == 'voxel':
                data[:, 1:] = coords
            else:
                data[:, 1:] = opts.transformCoords(coords, 'voxel', 'world')

        # save the file
        errTitle = strings.titles[  self, 'exportError']
        errMsg   = strings.messages[self, 'exportError']
        with status.reportIfError(errTitle, errMsg):
            np.savetxt(filename, data, fmt='%0.8f')


    @actions.action
    def screenshot(self):
        """Creates and runs a :class:`.ScreenshotAction`, which propmts the
        user to save the plot to a file.
        """
        screenshot.ScreenshotAction(self.overlayList,
                                    self.displayCtx,
                                    self.__canvas)()


    def __bindToDataSeries(self, ds, bind=True):
        """Binds/unbinds the GUI widgets (:attr:`colour`, :attr:`resolution`),
        etc to/from the given :class:`SampleLineDataSeries` instance.
        """
        ds.bindProps('colour',     self, unbind=not bind)
        ds.bindProps('lineWidth',  self, unbind=not bind)
        ds.bindProps('lineStyle',  self, unbind=not bind)
        ds.bindProps('resolution', self, unbind=not bind)
        ds.bindProps('interp',     self, unbind=not bind)
        ds.bindProps('normalise',  self, unbind=not bind)


    def __updateInfo(self):
        """Called while the mouse is being dragged. Updates the information
        about the sampling line (e.g. start/end voxel coordinates) that is
        displayed at the top of the ``SampleLinePanel``.
        """
        start = self.__profile.sampleStart
        end   = self.__profile.sampleEnd
        if start is None or end is None:
            self.__vfromval.SetLabel('')
            self.__vtoval  .SetLabel('')
            self.__wfromval.SetLabel('')
            self.__wtoval  .SetLabel('')
            self.__lenval  .SetLabel('')
            return

        overlay = self.displayCtx.getSelectedOverlay()
        opts    = self.displayCtx.getOpts(overlay)
        image   = opts.referenceImage

        # display world and voxel coordinates
        if image is not None:
            opts   = self.displayCtx.getOpts(image)
            units  = image.xyzUnits
            units  = strings.nifti.get(('xyz_unit', units), '(unknown units)')
            ws     = opts.transformCoords(start, 'display', 'world')
            we     = opts.transformCoords(end,   'display', 'world')
            vs     = opts.transformCoords(start, 'display', 'voxel')
            ve     = opts.transformCoords(end,   'display', 'voxel')
            length = affine.veclength(ws - we)[0]

            vs     = '[{:.2f}, {:.2f}, {:.2f}]'.format(*vs)
            ve     = '[{:.2f}, {:.2f}, {:.2f}]'.format(*ve)
            ws     = '[{:.2f}, {:.2f}, {:.2f}]'.format(*ws)
            we     = '[{:.2f}, {:.2f}, {:.2f}]'.format(*we)
            length = f'{length:.2f} {units}'

            self.__vfromval.SetLabel(vs)
            self.__vtoval  .SetLabel(ve)
            self.__wfromval.SetLabel(ws)
            self.__wtoval  .SetLabel(we)
            self.__lenval  .SetLabel(length)


    def __onMouseDown(self, *a):
        """Called on mouse down events on an :class:`.OrthoPanel` canvas.
        Calls :meth:`__updateInfo`.
        """
        self.__updateInfo()


    def __onMouseDrag(self, *a):
        """Called on mouse drag events on an :class:`.OrthoPanel` canvas.
        Calls :meth:`__updateInfo`.
        """
        self.__updateInfo()


    def __onMouseUp(self, *a):
        """Called on mouse up events on an :class:`.OrthoPanel` canvas.
        Samples and plots data from the currently selected overlay along
        the drawn sample line.
        """

        start  = self.__profile.sampleStart
        end    = self.__profile.sampleEnd
        image  = self.displayCtx.getSelectedOverlay()

        if start is None or end is None:
            return

        if image is None or not isinstance(image, fslimage.Image):
            return

        if self.__current is not None:
            self.__bindToDataSeries(self.__current, False)

        # Round to avoid floating point imprecision
        opts   = self.displayCtx.getOpts(image)
        vstart = opts.transformCoords(start, 'display', 'voxel').round(6)
        vend   = opts.transformCoords(end,   'display', 'voxel').round(6)

        series = SampleLineDataSeries(image,
                                      self.overlayList,
                                      self.displayCtx,
                                      self.__canvas,
                                      opts.index(),
                                      vstart,
                                      vend)

        self.__bindToDataSeries(series)
        self.__current = series
        self.__canvas.asyncDraw()


    def __draw(self):
        """Passed as the ``drawFunc`` to the :class:`.PlotCanvas`. Calls
        :meth:`.PlotCanvas.drawDataSeries`.
        """
        if self.__current is None: extras = None
        else:                      extras = [self.__current]
        self.__canvas.drawDataSeries(extraSeries=extras, refresh=True)


class ExportSampledDataDialog(wx.Dialog):
    """The ``ExportSampledDataDialog`` is used by the
    :meth:`SampleLinePanel.export` method to ask the user which sample data
    they want to save, and whether they want to save the sample point
    coordinates to file as well as the samples themselves.
    """


    def __init__(self, parent, series):
        """Create an ``ExportSampledDataDialog``.

        :arg parent: ``wx`` parent object
        :arg series: Sequence of ``SampleLineDataSeries`` instances. Must
                     contain at least one series. If there is more than one
                     series, the user is asked to choose one.
        """
        title        = strings.titles[self]
        coords       = strings.choices[self, 'saveCoordinates']
        seriesLabels = [s.label for s in series]
        coordsLabels = list(coords.values())
        coords       = list(coords.keys())

        wx.Dialog.__init__(self,
                           parent,
                           title=title,
                           style=wx.DEFAULT_DIALOG_STYLE)

        self.__series       = series
        self.__coords       = coords
        self.__coordsChoice = wx.Choice(self, choices=coordsLabels)
        self.__coordsChoice.SetSelection(0)

        coordsLabel = wx.StaticText(self)
        ok          = wx.Button(self, id=wx.ID_OK)
        cancel      = wx.Button(self, id=wx.ID_CANCEL)

        coordsLabel.SetLabel(strings.labels[self, 'coords'])
        ok         .SetLabel(strings.labels[self, 'ok'])
        cancel     .SetLabel(strings.labels[self, 'cancel'])

        if len(series) > 1:
            self.__seriesChoice = wx.Choice(self, choices=seriesLabels)
            self.__seriesChoice.SetSelection(0)
            seriesLabel = wx.StaticText(self)
            seriesLabel.SetLabel(strings.labels[self, 'series'])

        ok.SetDefault()
        ok    .Bind(wx.EVT_BUTTON, self.__onOk)
        cancel.Bind(wx.EVT_BUTTON, self.__onCancel)

        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        mainSizer   = wx.BoxSizer(wx.VERTICAL)

        buttonSizer.Add((10, 10), proportion=1)
        buttonSizer.Add(ok)
        buttonSizer.Add((10, 10))
        buttonSizer.Add(cancel)
        buttonSizer.Add((10, 10), proportion=1)

        mainSizer.Add((10, 10), proportion=1)

        if len(series) > 1:
            mainSizer.Add(seriesLabel,
                          flag=wx.ALIGN_CENTRE | wx.ALL, border=10)
            mainSizer.Add((10, 10))
            mainSizer.Add(self.__seriesChoice,
                          flag=wx.ALIGN_CENTRE | wx.ALL, border=10)
            mainSizer.Add((10, 10))

        mainSizer.Add(coordsLabel, flag=wx.ALIGN_CENTRE | wx.ALL, border=10)
        mainSizer.Add((10, 10))
        mainSizer.Add(self.__coordsChoice,
                      flag=wx.ALIGN_CENTRE | wx.ALL, border=10)
        mainSizer.Add((10, 10), proportion=1)
        mainSizer.Add(buttonSizer, flag=wx.ALIGN_CENTRE | wx.ALL, border=10)
        mainSizer.Add((10, 10))

        self.SetSizer(mainSizer)
        self.Layout()
        self.Fit()
        self.CentreOnParent()


    def __onOk(self, ev):
        """Called when the ok button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_OK)


    def __onCancel(self, ev):
        """Called when the cancel button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_CANCEL)


    def GetCoordinates(self):
        """Return one of ``'none'``, ``'voxel'``, or ``'world'``, denoting
        the user's preference for saving coordinates to the file.
        """
        idx = self.__coordsChoice.GetSelection()
        return self.__coords[idx]


    def GetSeries(self):
        """Return the :class:`.SampleLineDataSeries` that was selected."""
        if len(self.__series) == 1:
            return self.__series[0]
        idx = self.__seriesChoice.GetSelection()
        return self.__series[idx]



#!/usr/bin/env python
#
# saveflirtxfm.py - The SaveFlirtXfmAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SaveFlirtXfmAction` class, an
:class:`.Action` which allows the user to save an :class:`.Image`
transformation to disk for use with FLIRT.
"""


import logging

import numpy as np

import fsl.data.image               as fslimage
import fsl.transform.affine         as affine
import fsl.transform.flirt          as flirt
import fsleyes_widgets.utils.status as status
import fsleyes.strings              as strings
import fsleyes.actions.base         as base
from . import                          applyflirtxfm


log = logging.getLogger(__name__)


class SaveFlirtXfmAction(base.NeedOverlayAction):
    """The :class:`SaveFlirtXfmAction` class is an :class:`.Action` which
    allows the user to save an :class:`.Image` transformation to disk, either
    as a FLIRT matrix, or a voxel-to-world matrix.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``SaveFlirtXfmAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.NeedOverlayAction.__init__(
            self, overlayList, displayCtx, func=self.__saveFlirtXfm)
        self.__frame = frame


    def destroy(self):
        """Must be called when this ``SaveFlirtXfmAction`` is no longer needed.
        """
        self.__frame = None
        base.NeedOverlayAction.destroy(self)


    def __saveFlirtXfm(self):
        """Called when this action is executed. Prompts the user to save
        a FLIRT transform for the currently selected image.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        overlay     = displayCtx.getSelectedOverlay()

        affType, matFile, refFile = applyflirtxfm.promptForFlirtFiles(
            self.__frame, overlay, overlayList, displayCtx, save=True)

        if all((affType is None, matFile is None, refFile is None)):
            return

        if affType == 'flirt':
            xform = calculateTransform(
                overlay, overlayList, displayCtx, refFile)
        else:
            xform = overlay.voxToWorldMat

        errtitle = strings.titles[  self, 'error']
        errmsg   = strings.messages[self, 'error']

        with status.reportIfError(errtitle, errmsg):
            np.savetxt(matFile, xform, fmt='%0.10f')


def calculateTransform(overlay,
                       overlayList,
                       displayCtx,
                       refFile,
                       srcXform=None):
    """Calculates and returns FLIRT transformation from the given ``overlay``
    to the image specified by the given ``refFile``.
    """

    # The reference image might
    # already be in the overlay list.
    refImg = overlayList.find(refFile)

    if refImg is None:
        refImg = fslimage.Image(refFile, loadData=False)

    return flirt.sformToFlirtMatrix(overlay, refImg, srcXform)



#!/usr/bin/env python
#
# fsleyes_kernelmanager.py - Custom Jupyter notebook kernel manager.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


from notebook.services.kernels.kernelmanager import MappingKernelManager


class FSLeyesNotebookKernelManager(MappingKernelManager):
    """Custom jupter ``MappingKernelManager`` which forces every notebook
    to connect to the embedded FSLeyes IPython kernel.

    See https://github.com/ebanner/extipy
    """


    connfile = ''
    """Path to the IPython kernel connection file that all notebooks should
    connect to.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def __patch_connection(self, kernel):
        """Connects the given kernel to the IPython kernel specified by
        ``connfile``.
        """
        kernel.hb_port      = 0
        kernel.shell_port   = 0
        kernel.stdin_port   = 0
        kernel.iopub_port   = 0
        kernel.control_port = 0
        kernel.load_connection_file(self.connfile)


    async def start_kernel(self, **kwargs):
        """Overrides ``MappingKernelManager.start_kernel``. Connects
        all new kernels to the IPython kernel specified by ``connfile``.
        """
        kid    = await super().start_kernel(**kwargs)
        kernel = self._kernels[kid]
        self.__patch_connection(kernel)
        return kid


    def restart_kernel(self, *args, **kwargs):
        """Overrides ``MappingKernelManager.restart_kernel``. Does nothing. """
        pass



#!/usr/bin/env python
#
# about.py - The AboutAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.AboutAction` class, an action which
displays an about dialog for *FSLeyes*.
"""


from . import base

import fsleyes_widgets as fwidgets


class AboutAction(base.Action):
    """The ``AboutAction`` class is an action which displays an
    :class:`.AboutDialog`, containing information about *FSLeyes*.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create an ``AboutAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The master :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        base.Action.__init__(self, overlayList, displayCtx, self.__showDialog)
        self.__frame = frame


    def __showDialog(self):
        """Creates and shows an :class:`.AboutDialog`. """

        import fsleyes.about as aboutdlg

        dlg = aboutdlg.AboutDialog(self.__frame)
        dlg.Show()

        # When running over X11/SSH, CentreOnParent
        # causes the dialog to be moved way off.
        if not fwidgets.inSSHSession():
            dlg.CentreOnParent()



#!/usr/bin/env python
#
# glrgbvolume.py - The GLRGBVolume class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLRGBVolume` class, used for rendering
RGB(A) :class:`.Image` overlays.
"""


import numpy                as np
import OpenGL.GL            as gl

import fsl.utils.idle       as idle
import fsl.transform.affine as affine

import fsleyes.gl           as fslgl
import fsleyes.gl.routines  as glroutines
import fsleyes.gl.textures  as textures
import fsleyes.gl.resources as glresources
from . import                  glimageobject


class GLRGBVolume(glimageobject.GLImageObject):
    """The ``GLRGBVolume`` class is used to render RGB(A) :class:`.Image`
    overlays. The RGB(A) value at each voxel is directly used as the colour
    for that voxel.
    """


    def __init__(self, image, overlayList, displayCtx, canvas, threedee):
        """Create a ``GLRGBVolume``.

        :arg image:       The :class:`.Image` instance.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg canvas:      The canvas doing the drawing.
        :arg threedee:    2D or 3D rendering
        """
        glimageobject.GLImageObject.__init__(self,
                                             image,
                                             overlayList,
                                             displayCtx,
                                             canvas,
                                             threedee)

        self.shader       = None
        self.imageTexture = None

        self.addListeners()
        self.refreshImageTexture()

        def init():
            fslgl.glrgbvolume_funcs.init(self)
            self.notify()

        idle.idleWhen(init, self.textureReady)


    def destroy(self):
        """Must be called when this ``GLRGBVolume`` is no longer needed. """
        self.removeListeners()
        fslgl.glvolume_funcs.destroy(self)
        glresources.delete(self.imageTexture.name)
        glimageobject.GLImageObject.destroy(self)
        self.imageTexture = None


    def addListeners(self):
        """Adds listeners to :class:`.Display` and :class:`.VolumeRGBOpts`
        properties which should result in the display being refreshed.
        """
        display = self.display
        opts    = self.opts
        name    = self.name

        def shader(*a):
            self.updateShaderState()

        def notify(*a):
            self.notify()

        display.addListener('brightness',    name, shader, weak=False)
        display.addListener('contrast',      name, shader, weak=False)
        display.addListener('alpha',         name, shader, weak=False)
        opts   .addListener('interpolation', name, self._interpChanged)
        opts   .addListener('rColour',       name, shader, weak=False)
        opts   .addListener('gColour',       name, shader, weak=False)
        opts   .addListener('bColour',       name, shader, weak=False)
        opts   .addListener('suppressR',     name, shader, weak=False)
        opts   .addListener('suppressG',     name, shader, weak=False)
        opts   .addListener('suppressB',     name, shader, weak=False)
        opts   .addListener('suppressA',     name, shader, weak=False)
        opts   .addListener('suppressMode',  name, shader, weak=False)
        opts   .addListener('transform',     name, notify, weak=False)
        opts   .addListener('displayXform',  name, notify, weak=False)


    def removeListeners(self):
        """Removes the property listeners that were added in
        :meth:`addListeners`.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        display.removeListener('brightness',    name)
        display.removeListener('contrast',      name)
        display.removeListener('alpha',         name)
        opts   .removeListener('interpolation', name)
        opts   .removeListener('rColour',       name)
        opts   .removeListener('gColour',       name)
        opts   .removeListener('bColour',       name)
        opts   .removeListener('suppressR',     name)
        opts   .removeListener('suppressG',     name)
        opts   .removeListener('suppressB',     name)
        opts   .removeListener('suppressA',     name)
        opts   .removeListener('suppressMode',  name)
        opts   .removeListener('transform',     name)
        opts   .removeListener('displayXform',  name)


    def ready(self):
        """Returns ``True`` if this ``GLRGBVolume`` is ready to be used,
        ``False`` otherwise.
        """
        return (self.shader is not None) and self.textureReady()


    def textureReady(self):
        """Returns ``True`` if the image texture is ready to be used,
        ``False`` otherwise.
        """
        return ((self.imageTexture is not None) and
                (self.imageTexture.ready()))


    def updateShaderState(self, *args, **kwargs):
        """Calls :func:`.gl14.glrgbvolume_funcs.updateShaderState` or
        :func:`.gl12.glrgbvolume_funcs.updateShaderState`.
        """
        fslgl.glrgbvolume_funcs.updateShaderState(self)
        self.notify()


    def refreshImageTexture(self):
        """(Re-)creates an :class:`.ImageTexture` or :class:`.ImageTexture2D`
        to store the image data.
        """

        texName = '{}_{}' .format(type(self).__name__, id(self.image))
        nvals   = self.overlay.nvals

        if self.opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                                 interp = gl.GL_LINEAR

        if nvals == 1:
            nvals = self.overlay.shape[-1]

        self.imageTexture = glresources.get(
            texName,
            textures.createImageTexture,
            texName,
            self.overlay,
            nvals=nvals,
            interp=interp,
            notify=False)

        self.imageTexture.register(self.name, self.__imageTextureChanged)


    def _interpChanged(self, *a):
        """Called when the :attr:`.VolumeRGBOpts.interpolation` changes.
        Updates the image texture.
        """

        opts = self.opts

        if opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                            interp = gl.GL_LINEAR

        self.imageTexture.set(interp=interp)


    def __imageTextureChanged(self, *a):
        """Called when the ``imageTexture`` changes. Calls
        :meth:`updateShaderState`.
        """
        self.updateShaderState()


    def generateVertices2D(self, zpos, axes, bbox=None):
        """Overrides :meth:`.GLImageObject.generateVertices2D`.

        Appliies the :meth:`.ImageTextureBase.texCoordXform` to the texture
        coordinates - this is performed to support 2D images/textures.
        """

        vertices, voxCoords, texCoords = \
            glimageobject.GLImageObject.generateVertices2D(
                self, zpos, axes, bbox)

        texCoords = affine.transform(
            texCoords, self.imageTexture.texCoordXform(self.overlay.shape))

        return vertices, voxCoords, texCoords


    def channelColours(self):
        """Returns a ``numpy`` array of shape ``(3, 4)``, containing the
        colours to use for each of the three channels.
        """

        opts     = self.opts
        display  = self.display
        alpha    = display.alpha / 100

        colours        = np.zeros((3, 4), dtype=np.float32)
        colours[0, :3] = opts.rColour[:3]
        colours[1, :3] = opts.gColour[:3]
        colours[2, :3] = opts.bColour[:3]
        colours[:,  3] = alpha

        if   opts.suppressMode == 'white':       suppress = [1, 1, 1, alpha]
        elif opts.suppressMode == 'black':       suppress = [0, 0, 0, alpha]
        elif opts.suppressMode == 'transparent': suppress = [0, 0, 0, 0]

        if opts.suppressR: colours[0, :] = suppress
        if opts.suppressG: colours[1, :] = suppress
        if opts.suppressB: colours[2, :] = suppress

        return colours


    def preDraw(self, xform=None, bbox=None):
        """Called before a draw. Binds the image texture. """
        self.imageTexture.bindTexture(gl.GL_TEXTURE0)


    def draw2D(self, zpos, axes, xform=None, bbox=None):
        """Calls :func:`.glrgbvolume_funcs.draw2D`. """
        with glroutines.enabled((gl.GL_CULL_FACE)):
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
            gl.glCullFace(gl.GL_BACK)
            gl.glFrontFace(self.frontFace())
            fslgl.glrgbvolume_funcs.draw2D(self, zpos, axes, xform, bbox)


    def drawAll(self, axes, zposes, xforms):
        """Calls :func:`.glrgbvolume_funcs.drawAll`. """
        fslgl.glrgbvolume_funcs.drawAll(self, axes, zposes, xforms)


    def draw3D(self, xform=None, bbox=None):
        """Does nothing. """
        pass


    def postDraw(self, xform=None, bbox=None):
        """Called after a draw. Unbinds the image texture. """
        self.imageTexture.unbindTexture()



#!/usr/bin/env python
#
# texture.py - The Texture and Texture2D classes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Texture` class, which is the base classes
for all other FSLeyes texture types. See also the :class:`.Texture2D` and
:class:`.Texture3D` classes.
"""


import              logging
import              contextlib
import functools as ft

import numpy                        as np
import OpenGL.GL                    as gl

import fsl.utils.idle               as idle

import fsl.utils.notifier           as notifier
import fsl.transform.affine         as affine
import fsleyes_widgets.utils.status as status
import fsleyes.strings              as strings
from . import data                  as texdata


log = logging.getLogger(__name__)


class TextureBase(object):
    """Base mixin class used by the :class:`Texture` class.

    This class provides logic for texture lifecycle management
    (creation/destruction) and usage.


    .. autosummary::
       :nosignatures:

       name
       handle
       target
       ndim
       nvals
       isBound
       bound
       bindTexture
       unbindTexture


    The :meth:`bound` method (which uses :meth:`bindTexture` and
    :meth:`unbindTexture`) method allows you to bind a texture object to a GL
    texture unit. For example, let's say we have a texture object called
    ``tex``, and we want to configure and use it::

        import OpenGL.GL as gl

        # When we want to use the texture in a
        # scene render, we need to bind it to
        # a texture unit.
        with tex.bound(gl.GL_TEXTURE0):

            # use linear interpolation
            tex.interp = gl.GL_LINEAR

            # ...
            # Do the render
            # ...
    """


    def __init__(self, name, ndims, nvals):
        """Create a ``TextureBase``.

        :arg name:  The name of this texture - should be unique.
        :arg ndims: Number of dimensions - must be 1, 2 or 3.
        :arg nvals: Number of values stored in each texture element.
        """
        if   ndims == 1: ttype = gl.GL_TEXTURE_1D
        elif ndims == 2: ttype = gl.GL_TEXTURE_2D
        elif ndims == 3: ttype = gl.GL_TEXTURE_3D
        else:            raise ValueError('Invalid number of dimensions')

        self.__texture     = int(gl.glGenTextures(1))
        self.__ttype       = ttype
        self.__name        = name
        self.__ndims       = ndims
        self.__nvals       = nvals
        self.__bound       = 0
        self.__textureUnit = None


    def __del__(self):
        """Prints a log message."""

        # log might get deleted before us
        try:
            log.debug('%s.del (%s)', type(self).__name__, id(self))
        except Exception:
            pass


    def destroy(self):
        """Must be called when this ``TextureBase`` is no longer needed.
        Deletes the texture handle.
        """

        log.debug('Deleting %s (%s) for %s: %s',
                  type(self).__name__, id(self),
                  self.__name, self.__texture)

        gl.glDeleteTextures(self.__texture)
        self.__texture = None


    @property
    def name(self):
        """Returns the name of this texture. This is not the GL texture name,
        rather it is the unique name passed into :meth:`__init__`.
        """
        return self.__name


    @property
    def handle(self):
        """Returns the GL texture handle for this texture. """
        return self.__texture


    @property
    def target(self):
        """Returns the type of this texture - ``GL_TEXTURE_1D``,
        ``GL_TEXTURE_2D`` or ``GL_TEXTURE_3D``.
        """
        return self.__ttype


    @property
    def ndim(self):
        """Return the number of dimensions of this texture - 1, 2, or 3. """
        return self.__ndims


    @property
    def nvals(self):
        """Return the number of values stored at each point in this texture.
        """
        return self.__nvals


    def isBound(self):
        """Returns ``True`` if this texture is currently bound, ``False``
        otherwise.

        .. note:: This method assumes that the :meth:`bindTexture` and
                 :meth:`unbindTexture` methods are called in pairs.
        """
        return self.__bound > 0


    @contextlib.contextmanager
    def bound(self, textureUnit=None):
        """Context manager which can be used to bind and unbind this texture,
        instead of manually calling :meth:`bindTexture` and
        :meth:`unbindTexture`

        :arg textureUnit: The texture unit to bind this texture to, e.g.
                          ``GL_TEXTURE0``.
        """
        try:
            self.bindTexture(textureUnit)
            yield
        finally:
            self.unbindTexture()


    def bindTexture(self, textureUnit=None):
        """Activates and binds this texture.

        :arg textureUnit: The texture unit to bind this texture to, e.g.
                          ``GL_TEXTURE0``.
        """

        if self.__bound == 0:

            if textureUnit is not None:
                gl.glActiveTexture(textureUnit)

            gl.glBindTexture(self.__ttype, self.__texture)

            self.__textureUnit = textureUnit

        self.__bound += 1


    def unbindTexture(self):
        """Unbinds this texture. """

        if self.__bound == 1:
            if self.__textureUnit is not None:
                gl.glActiveTexture(self.__textureUnit)

            gl.glBindTexture(self.__ttype, 0)

            self.__textureUnit = None

        self.__bound = max(0, self.__bound - 1)


class TextureSettingsMixin(object):
    """Mixin class used by the :class:`Texture` class.


    This class provides methods to get/set various settings which can
    be used to manipulate the texture. All of the logic which uses
    these settings is in the ``Texture`` class.


    The following settings can be changed:

    .. autosummary::
       :nosignatures:

       interp
       prefilter
       prefilterRange
       normalise
       normaliseRange
       border
       scales
       resolution

    Additional settings can be added via the ``settings`` argument to
    :meth:`__init__`. All settings can be changed via the :meth:`update`
    method.
    """

    def __init__(self, settings=None):
        """Create a ``TextureSettingsMixin``.

        :arg settings: Sequence of additional settings to make available.
        """

        defaults = ['interp',
                    'prefilter', 'prefilterRange',
                    'normalise', 'normaliseRange',
                    'border', 'resolution', 'scales']

        if settings is None: settings = defaults
        else:                settings = defaults + list(settings)

        self.__settings = {s : None for s in settings}


    @property
    def interp(self):
        """Return the current texture interpolation setting - either
        ``GL_NEAREST`` or ``GL_LINEAR``.
        """
        return self.__settings['interp']


    @interp.setter
    def interp(self, interp):
        """Sets the texture interpolation. """
        self.update(interp=interp)


    @property
    def prefilter(self):
        """Return the current prefilter function - texture data is passed
        through this function before being uploaded to the GPU.

        If this function changes the range of the data, you must also
        provide a ``prefilterRange`` function - see :meth:`prefilterRange`.
        """
        return self.__settings['prefilter']


    @prefilter.setter
    def prefilter(self, prefilter):
        """Set the prefilter function """
        self.update(prefilter=prefilter)


    @property
    def prefilterRange(self):
        """Return the current prefilter range function - if the ``prefilter``
        function changes the data range, this function must be provided. It
        is passed two parameters - the known data minimum and maximum, and
        must adjust these values so that they reflect the adjusted range of
        the data that was passed to the ``prefilter`` function.
        """
        return self.__settings['prefilterRange']


    @prefilterRange.setter
    def prefilterRange(self, prefilterRange):
        """Set the prefilter range function. """
        self.update(prefilter=prefilterRange)


    @property
    def normalise(self):
        """Return the current normalisation state.

        If ``normalise=True``, the data is normalised to lie in the range
        ``[0, 1]`` (or normalised to the full range, if being stored as
        integers) before being stored. The data is normalised according to
        the minimum/maximum of the data, or to a normalise range set via
        the :meth:`normaliseRange`.

        Set this to ``False`` to disable normalisation.

        .. note:: If the data is not of a type that can be stored natively
                  as a texture, the data is automatically normalised,
                  regardless of the value specified here.
        """
        return self.__settings['normalise']


    @normalise.setter
    def normalise(self, normalise):
        """Enable/disable normalisation. """
        self.update(normalise=normalise)


    @property
    def normaliseRange(self):
        """Return the current normalise range.

        If normalisation is enabled (see :meth:`normalise`), or necessary,
        the data is normalised according to either its minimum/maximum, or
        to the range specified via this method.

        This parameter must be a sequence of two values, containing the
        ``(min, max)`` normalisation range. The data is then normalised to
        lie in the range ``[0, 1]`` (or normalised to the full range, if being
        stored as integers) before being stored.

        If ``None``, the data minimum/maximum are calculated and used.
        """
        return self.__settings['normaliseRange']


    @normaliseRange.setter
    def normaliseRange(self, normaliseRange):
        """Set the normalise range. """
        self.update(normaliseRange=normaliseRange)


    @property
    def border(self):
        """Return the texture border colour. Set this to a tuple of four values
        in the range 0 to 1, or ``None`` for no border (in which case the
        texture coordinates will be clamped to edges).
        """
        return self.__settings['border']


    @border.setter
    def border(self, border):
        """Return the texture border colour."""
        self.update(border=border)


    @property
    def scales(self):
        """Return the scaling factors for each axis of the texture data.

        These values are solely used to calculate the sub-sampling rate if the
        resolution (as set by :meth:`resolution`) is in terms of something
        other than data indices (e.g. :class:`.Image` pixdims).
        """
        return self.__settings['scales']


    @scales.setter
    def scales(self, scales):
        """Set the texture data axis scaling factors. """
        self.update(scales=scales)


    @property
    def resolution(self):
        """Return the current texture data resolution - this value is passed
        to the :func:`.routines.subsample` function, in the
        :func:`.prepareData` function.
        """
        return self.__settings['resolution']


    @resolution.setter
    def resolution(self, resolution):
        """Set the texture data resolution. """
        self.update(resolution=resolution)


    def update(self, **kwargs):
        """Set any parameters on this ``TextureSettingsMixin``. Valid keyword
        arguments are:

        ================== ==========================
        ``interp``         See :meth:`interp`.
        ``prefilter``      See :meth:`prefilter`.
        ``prefilterRange`` See :meth:`prefilterRange`
        ``normalise``      See :meth:`normalise.`
        ``normaliseRange`` See :meth:`normaliseRange`
        ``border``         See :meth:`border`
        ``scales``         See :meth:`scales`.
        ``resolution``     See :meth:`resolution`
        ================== ==========================

        :returns: A ``dict`` of ``{attr : changed}`` mappings, indicating
                  which properties have changed value.
        """

        changed = {}
        for s in self.__settings.keys():
            oldval             = self.__settings[s]
            newval             = kwargs.get(s, self.__settings[s])
            changed[s]         = oldval != newval
            self.__settings[s] = newval
        return changed


class Texture(notifier.Notifier, TextureBase, TextureSettingsMixin):
    """The ``Texture`` class is the base class for all other texture types in
    *FSLeyes*. This class is not intended to be used directly - use one of the
    sub-classes instead.


    A texture can be bound and unbound via the methods of the
    :class:`TextureBase` class. Various texture settings can be changed
    via the methods of the :class:`TextureSettingsMixin` class. In the majority
    of cases, in order to draw or configure a texture, it needs to be bound
    (although this depends on the sub-class).


    In order to use a texture, at the very least you need to provide some
    data, or specify a type and shape. This can be done either via the
    :meth:`data`/:meth:`shape`/:meth:`dtype` methods, or by the :meth:`set`
    method. If you specify a shape and data type, any previously specified
    data will be lost, and vice versa.


    Calling :meth:`set` will usually cause the texture to be reconfigured and
    refreshed, although you can also force a refresh by calling the
    :meth:`refresh` method directly.


    The following properties can be queried to retrieve information about the
    tetxure; some will return ``None`` until you have provided some data (or
    a shape and type):

    .. autosummary::
       :nosignatures:

       voxValXform
       invVoxValXform
       shape
       dtype
       textureType
       baseFormat
       internalFormat
       data
       preparedData


    When a ``Texture`` is created, and when its settings are changed, it may
    need to prepare the data to be passed to OpenGL - for large textures, this
    can be a time consuming process, so this may be performed on a separate
    thread using the :mod:`.idle` module (unless the ``threaded`` parameter to
    :meth:`__init__` is set to ``False``). The :meth:`ready` method returns
    ``True`` or ``False`` to indicate whether the ``Texture`` is ready to be
    used.


    Furthermore, the ``Texture`` class derives from :class:`.Notifier`, so
    listeners can register to be notified when an ``Texture`` is ready to
    be used.


    For textures with multiple values per voxel, it is assumed that these
    values are indexed with the first dimension of the texture data (as passed
    to :meth:`data` or :meth:`set`).


    ``Texture`` sub-classes (e.g. :class:`.Texture2D`, :class:`.Texture3D`,
    :class:`.ColourMapTexture`) must override the :meth:`doRefresh` method
    such that it performs the GL calls required to configure the textureb.


    See the :mod:`.resources` module for a method of sharing texture resources.
    """


    def __init__(self,
                 name,
                 ndims,
                 nvals,
                 threaded=False,
                 settings=None,
                 textureFormat=None,
                 internalFormat=None,
                 **kwargs):
        """Create a ``Texture``.

        :arg name:           The name of this texture - should be unique.

        :arg ndims:          Number of dimensions - must be 1, 2 or 3.

        :arg nvals:          Number of values stored in each texture element.

        :arg threaded:       If ``True``, the texture data will be prepared on
                             a separate thread (on calls to
                             :meth:`refresh`). If ``False``, the texture data
                             is prepared on the calling thread, and the
                             :meth:`refresh` call will block until it has been
                             prepared.

        :arg settings:       Additional settings to make available through the
                             :class:`TextureSettingsMixin`.

        :arg textureFormat:  Texture format to use - if not specified, this is
                             automatically determined. If specified, an
                             ``internalFormat`` must also be specified.

        :arg internalFormat: Internal texture format to use - if not specified,
                             this is automatically determined.

        All other arguments are passed through to the initial call to
        :meth:`set`.

        .. note:: All subclasses must accept a ``name`` as the first parameter
                  to their ``__init__`` method, and must pass said ``name``
                  through to the :meth:`__init__` method.

        .. note:: In normal cases, the ``textureFormat`` and ``internalFormat``
                  do not need to be specified - they will be automatically
                  determined using the :func:`.data.getTextureType` function.
                  However, there can be instances where a specific texture type
                  needs to be used. In these instances, it is up to the calling
                  code to ensure that the texture data can be coerced into
                  the correct GL data type.
        """

        TextureBase         .__init__(self, name, ndims, nvals)
        TextureSettingsMixin.__init__(self, settings)

        if ((textureFormat is not None) and (internalFormat is     None)) or \
           ((textureFormat is     None) and (internalFormat is not None)):
            raise ValueError('Both textureFormat and internalFormat '
                             'must be specified')

        self.__ready    = False
        self.__threaded = threaded

        # The data, type and shape are
        # refreshed on every call to
        # set or refresh (the former
        # calls the latter)
        self.__data         = None
        self.__dtype        = None
        self.__shape        = None
        self.__preparedData = None

        # The data is refreshed on
        # every call to set or refresh
        # These attributes are set by
        # the __determineTextureType
        # and __prepareTextureData
        # methods (which are called
        # by refresh)
        self.__voxValXform    = None
        self.__invVoxValXform = None

        self.__autoTexFmt     = textureFormat is None
        self.__texFmt         = textureFormat
        self.__texIntFmt      = internalFormat
        self.__texDtype       = None

        # If threading is enabled, texture
        # refreshes are performed with an
        # idle.TaskThread.
        if threaded:
            self.__taskThread = idle.TaskThread()
            self.__taskName   = '{}_{}_refresh'.format(type(self).__name__,
                                                       id(self))

            self.__taskThread.daemon = True
            self.__taskThread.start()
        else:
            self.__taskThread = None
            self.__taskName   = None

        self.set(**kwargs)


    def destroy(self):
        """Must be called when this ``Texture`` is no longer needed.
        """
        TextureBase.destroy(self)

        if self.__taskThread is not None:
            self.__taskThread.stop()

        self.__taskThread   = None
        self.__data         = None
        self.__preparedData = None



    def ready(self):
        """Returns ``True`` if this ``Texture`` is ready to be used,
        ``False`` otherwise.
        """
        return self.__ready


    @property
    def voxValXform(self):
        """Return a transformation matrix that can be used to transform
        values read from the texture back to the original data range.
        """
        return self.__voxValXform


    @property
    def invVoxValXform(self):
        """Return a transformation matrix that can be used to transform
        values in the original data range to values as read from the texture.
        """
        return self.__invVoxValXform


    def texCoordXform(self, origShape):
        """Returns a transformation matrix which can be used to adjust a set of
        3D texture coordinates so they can index the underlying texture, which
        may be 2D.

        This implementation returns an identity matrix, but it is overridden
        by the .Texture2D sub-class, which is sometimes used to store 3D image
        data.

        :arg origShape: Original data shape.
        """
        return np.eye(4)


    def invTexCoordXform(self, origShape):
        """Returns the inverse of :meth:`texCoordXform`. """
        return affine.invert(self.texCoordXform(origShape))


    @property
    def shape(self):
        """Return a tuple containing the texture data shape. """
        return self.__shape


    @shape.setter
    def shape(self, shape):
        """Set the texture data shape. """
        return self.set(shape=shape)


    @property
    def dtype(self):
        """Return the ``numpy`` data type of the texture data."""
        return self.__dtype


    @dtype.setter
    def dtype(self, dtype):
        """Set the ``numpy`` data type for the texture data."""
        self.set(dtype=dtype)


    @property
    def textureType(self):
        """Return the texture data type, e.g. ``gl.GL_UNSIGNED_BYTE``. """
        return self.__texDtype


    @property
    def baseFormat(self):
        """Return the base texture format, e.g. ``gl.GL_ALPHA``. """
        return self.__texFmt


    @property
    def internalFormat(self):
        """Return the sized/internal texture format, e.g. ``gl.GL_ALPHA8``. """
        return self.__texIntFmt


    @property
    def data(self):
        """Returns the data that has been passed to the :meth:`set` method. """
        return self.__data


    @data.setter
    def data(self, data):
        """Set the texture data - this get passed through to :meth:`set`. """
        self.set(data=data)


    @property
    def preparedData(self):
        """Returns the prepared data, i.e. the data as it has been copied
        to the GPU.
        """
        return self.__preparedData


    def shapeData(self, data, oldShape=None):
        """Shape the data so that it is ready for use as texture data.

        This implementation returns the data unchanged, but it is overridden
        by the ``Texture2D`` class, which is sometimes used to store 3D image
        data.

        :arg data:     ``numpy`` array containing the data to be shaped
        :arg oldShape: Original data shape, if this is a sub-array. If not
                       provided, taken from ``data``.
        """
        return data


    def set(self, **kwargs):
        """Set any parameters on this ``Texture``. Valid keyword arguments are:

        ================== ==============================================
        ``interp``         See :meth:`.interp`.
        ``data``           See :meth:`.data`.
        ``shape``          See :meth:`.shape`.
        ``dtype``          See :meth:`.dtype`.
        ``prefilter``      See :meth:`.prefilter`.
        ``prefilterRange`` See :meth:`.prefilterRange`
        ``normalise``      See :meth:`.normalise.`
        ``normaliseRange`` See :meth:`.normaliseRange`.
        ``scales``         See :meth:`.scales`.
        ``resolution``     See :meth:`.resolution`.
        ``refresh``        If ``True`` (the default), the :meth:`refresh`
                           function is called (but only if a setting has
                           changed).
        ``callback``       Optional function which will be called (via
                           :func:`.idle.idle`) when the texture has been
                           refreshed. Only called if ``refresh`` is
                           ``True``, and a setting has changed.
        ``notify``         Passed through to the :meth:`refresh` method.
        ================== ==============================================

        :returns: ``True`` if any settings have changed and the
                  ``Texture`` is being/needs to be refreshed, ``False``
                  otherwise.
        """

        changed  = TextureSettingsMixin.update(self, **kwargs)
        data     = kwargs.get('data',     None)
        shape    = kwargs.get('shape',    self.shape)
        dtype    = kwargs.get('dtype',    self.dtype)
        refresh  = kwargs.get('refresh',  True)
        notify   = kwargs.get('notify',   True)
        callback = kwargs.get('callback', None)

        changed['data']  = data  is not None
        changed['shape'] = shape != self.shape
        changed['dtype'] = dtype != self.dtype

        if not any(changed.values()):
            return False

        if data is not None:

            # The dtype attribute is set
            # later in __prepareTextureData,
            # as it may be different from
            # the dtype of the passed-in
            # data
            self.__data  = data
            self.__dtype = None
            dtype        = data.dtype

            # The first dimension is assumed to contain the
            # values, for multi-valued (e.g. RGB) textures
            if self.nvals > 1: self.__shape = data.shape[1:]
            else:              self.__shape = data.shape

        # If the data is of a type which cannot
        # be stored natively as an OpenGL texture,
        # and we don't have support for floating
        # point textures, the data must be
        # normalised. See determineType and
        # prepareData in the data module
        self.normalise = self.normalise or \
            (not texdata.canUseFloatTextures()[0] and
             (dtype not in (np.uint8, np.int8, np.uint16, np.int16)))

        # If the caller has not provided
        # a normalisation range, we have
        # to calculate it.
        if (data is not None) and \
           self.normalise and \
           (self.normaliseRange is None):
            self.normaliseRange = np.nanmin(data), np.nanmax(data)
            log.debug('Calculated %s data range for normalisation: '
                      '[%s - %s]', self.name, *self.normaliseRange)

        elif changed['shape'] or changed['dtype']:
            self.__data  = None
            self.__dtype = dtype
            self.__shape = shape

        refreshData = any((changed['data'],
                           changed['prefilter'],
                           changed['prefilterRange'],
                           changed['normaliseRange'] and self.normalise,
                           changed['resolution'],
                           changed['scales'],
                           changed['normalise']))

        if refresh:
            self.refresh(refreshData=refreshData,
                         notify=notify,
                         callback=callback)

        return True


    def refresh(self, refreshData=True, notify=True, callback=None):
        """(Re-)configures the OpenGL texture.

        :arg refreshData:  If ``True`` (the default), the texture data is
                           refreshed.

        :arg notify:       If ``True`` (the default), a notification is
                           triggered via the :class:`.Notifier` base-class,
                           when this ``Texture3D`` has been refreshed, and
                           is ready to use. Otherwise, the notification is
                           suppressed.

        :arg callback:     Optional function which will be called (via
                           :func:`.idle.idle`) when the texture has been
                           refreshed. Only called if ``refresh`` is
                           ``True``, and a setting has changed.

        .. note:: The texture data may be generated on a separate thread, using
                  the :func:`.idle.run` function. This is controlled by the
                  ``threaded`` parameter,  passed to :meth:`__init__`.
        """

        # Don't bother if data
        # or shape/type hasn't
        # been set
        data  = self.__data
        shape = self.__shape
        dtype = self.__dtype

        # We either need some data, or
        # we need a shape and data type.
        if data is None and (shape is None or dtype is None):
            return

        refreshData  = refreshData and (data is not None)
        self.__ready = False

        # This can take a long time for big
        # data, so we do it in a separate
        # thread using the idle module.
        def genData():

            # Another genData function is
            # already queued - don't run.
            # The TaskThreadVeto error
            # will stop the TaskThread from
            # calling configTexture as well.
            if self.__taskThread is not None and \
               self.__taskThread.isQueued(self.__taskName):
                raise idle.TaskThreadVeto()

            self.__determineTextureType()

            if refreshData:
                self.__prepareTextureData()

        # Once genData is finished, we pass the
        # result (see __prepareTextureData) to
        # the sub-class doRefresh method.
        def doRefresh():

            self.doRefresh()

            self.__ready = True

            if notify:
                self.notify()
            if callback is not None:
                callback()

        # Wrap the above functions in a report
        # decorator in case an error occurs
        title     = strings.messages[self, 'dataError']
        msg       = strings.messages[self, 'dataError']

        # the genData function is called on a separate thread,
        # but doRefresh is called on the idle/mainloop. So we
        # can use the reportErrorDecorator for the latter, but
        # not the former.
        doRefresh    = status.reportErrorDecorator(title, msg)(doRefresh)
        genDataError = ft.partial(status.reportError, title, msg)

        # Run asynchronously if we are
        # threaded, and we have data to
        # prepare - if we don't have
        # data, we run genData on the
        # current thread, because it
        # shouldn't do anything
        if self.__threaded and (data is not None):

            # TODO the task is already queued,
            #      but a callback function has been
            #      specified, should you queue the
            #      callback function?

            # Don't queue the texture
            # refresh task twice
            if not self.__taskThread.isQueued(self.__taskName):
                self.__taskThread.enqueue(genData,
                                          taskName=self.__taskName,
                                          onFinish=doRefresh,
                                          onError=genDataError)

        else:
            genData()
            doRefresh()


    def patchData(self, data, offset):
        """This is a shortcut method which can be used to replace part
        of the image texture data without having to regenerate the entire
        texture.

        The :meth:`set` and :meth:`refresh` methods are quite heavyweight, and
        are written in such a way that partial texture updates are not
        possible. This method allows small parts of the image texture to be
        quickly updated.
        """
        data = np.asarray(data)

        if len(data.shape) < self.ndim:
            newshape = list(data.shape) + [1] * (self.ndim - len(data.shape))
            data     = data.reshape(newshape)

        data = texdata.prepareData(
            data,
            prefilter=self.prefilter,
            prefilterRange=self.prefilterRange,
            resolution=self.resolution,
            scales=self.scales,
            normalise=self.normalise,
            normaliseRange=self.normaliseRange)[0]

        self.doPatch(data, offset)

        self.notify()


    def doRefresh(self):
        """Must be overridden by sub-classes to configure the texture.

        This method is not intended to be called externally - call
        :meth:`refresh` instead.

        This method should use the :meth:`preparedData`, or the :meth:`shape`,
        to configure the texture. Sub-classes can assume that at least one
        of these will not be ``None``.

        If ``preparedData`` is not ``None``, the ``shape`` should be ignored,
        and inferred from ``preparedData``.
        """
        raise NotImplementedError('Must be implemented by subclasses')


    def doPatch(self, data, offset):
        """Must be overridden by sub-classes to quickly update part of
        the texture data.

        This method is not intended to be called externally - call
        :meth:`patchData` instead.
        """
        raise NotImplementedError('Must be implemented by subclasses')


    def __determineTextureType(self):
        """Figures out how the texture data should be stored as an OpenGL
        texture. See the :func:`.data.getTextureType` function.

        This method sets the following attributes on this ``Texture`` instance:

        ==================== ==============================================
        ``__texFmt``         The texture format (e.g. ``GL_RGB``,
                             ``GL_LUMINANCE``, etc).

        ``__texIntFmt``      The internal texture format used by OpenGL for
                             storage (e.g. ``GL_RGB16``, ``GL_LUMINANCE8``,
                             etc).

        ``__texDtype``       The raw type of the texture data (e.g.
                             ``GL_UNSIGNED_SHORT``)
        ==================== ==============================================
        """

        if self.nvals not in (1, 3, 4):
            raise ValueError('Cannot create texture representation for {} '
                             '(nvals: {})'.format(self.dtype, self.nvals))

        if self.__data is None: dtype = self.__dtype
        else:                   dtype = self.__data.dtype

        normalise                = self.normalise
        nvals                    = self.nvals
        texDtype, texFmt, intFmt = texdata.getTextureType(
            normalise, dtype, nvals)

        if not self.__autoTexFmt:
            texFmt = self.__texFmt
            intFmt = self.__texIntFmt

        log.debug('Texture (%s) is to be stored as %s/%s/%s '
                  '(normalised: %s)',
                  self.name,
                  texdata.GL_TYPE_NAMES[texDtype],
                  texdata.GL_TYPE_NAMES[texFmt],
                  texdata.GL_TYPE_NAMES[intFmt],
                  normalise)

        self.__texFmt    = texFmt
        self.__texIntFmt = intFmt
        self.__texDtype  = texDtype


    def __prepareTextureData(self):
        """Prepare the texture data.

        This method passes the stored data to the :func:`.data.prepareData`
        function and then stores references to its return valuesa as
        attributes on this ``Texture`` instance:

        ==================== =============================================
        ``__preparedata``    A ``numpy`` array containing the image data,
                             ready to be copied to the GPU.

        ``__voxValXform``    An affine transformation matrix which encodes
                             an offset and a scale, which may be used to
                             transform the texture data from the range
                             ``[0.0, 1.0]`` to its raw data range.

        ``__invVoxValXform`` Inverse of ``voxValXform``.
        ==================== =============================================
        """

        data, voxValXform, invVoxValXform = texdata.prepareData(
            self.__data,
            prefilter=self.prefilter,
            prefilterRange=self.prefilterRange,
            resolution=self.resolution,
            scales=self.scales,
            normalise=self.normalise,
            normaliseRange=self.normaliseRange)

        self.__preparedData   = data
        self.__dtype          = data.dtype
        self.__voxValXform    = voxValXform
        self.__invVoxValXform = invVoxValXform



#!/usr/bin/env python
#
# autodisplay.py - Routines for configuring default overlay display
#                  settings.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :func:`audoDisplay` function, which is used
for automatically configuring overlay display settings.

The :autoDisplay` function is called when *FSLeyes* is started, and when
new overlays are loaded.
"""


import re
import sys
import logging
import os.path as op

import numpy as np

import fsl.data.image as fslimage


log = logging.getLogger(__name__)


def autoDisplay(overlay, overlayList, displayCtx, **kwargs):
    """Automatically configure display settings for the given overlay.

    :arg overlay:     The overlay object (e.g. an :class:`.Image` instance).
    :arg overlayList: The :class:`.OverlayList`.
    :arg displayCtx:  The :class:`.DisplayContext`.
    :arg kwargs:      Passed through to the overlay-type specific function.
    """

    oType = type(overlay).__name__
    func  = getattr(sys.modules[__name__], '_{}Display'.format(oType), None)

    if func is None:
        log.warn('Unknown overlay type: {}'.format(oType))
        return

    log.debug('Applying default display arguments for {}'.format(overlay))
    func(overlay, overlayList, displayCtx, **kwargs)


def _ImageDisplay(overlay, overlayList, displayCtx, **kwargs):
    """Configure default display settings for the given :class:`.Image`
    overlay.
    """

    if _isStatImage(overlay):
        _statImageDisplay(overlay, overlayList, displayCtx, **kwargs)
    elif _isPEImage(overlay):
        _peImageDisplay(  overlay, overlayList, displayCtx, **kwargs)

    # Automatically configure nice display range?


def _isStatImage(overlay):
    """Returns ``True`` if the given :class:`.Image` overlay looks like a
    statistic image, ``False`` otherwise.
    """

    basename = op.basename(overlay.dataSource)
    basename = fslimage.removeExt(basename)
    tokens   = ['zstat', 'tstat', 'fstat', 'zfstat']
    pattern  = r'({})\d+'.format('|'.join(tokens))

    return re.search(pattern, basename) is not None


def _isPEImage(overlay):
    """Returns ``True`` if the given :class:`.Image` overlay looks like a
    statistic image, ``False`` otherwise.
    """
    basename = op.basename(overlay.dataSource)
    basename = fslimage.removeExt(basename)
    tokens   = ['cope', 'pe']
    pattern  = r'^({})\d+'.format('|'.join(tokens))

    return re.search(pattern, basename) is not None


def _statImageDisplay(overlay,
                      overlayList,
                      displayCtx,
                      zthres=3.0,
                      posCmap=None,
                      negCmap=None):
    """Configure default display settings for the given statistic
    :class:`.Image` overlay.
    """

    opts        = displayCtx.getOpts(overlay)
    basename    = op.basename(overlay.dataSource)
    basename    = fslimage.removeExt(basename)

    pTokens     = ['p', 'corrp']
    statTokens  = ['zstat', 'tstat', 'zfstat']
    fStatTokens = ['fstat']

    # Rendered stat images (e.g.
    # rendered_thres_zstat1) are
    # generated specifically for
    # use with the Render1 colour
    # map.
    if 'rendered' in basename:
        opts.cmap = 'Render1'

    # Give each normal stat image
    # a different colour map
    else:
        cmap = _statImageDisplay.cmaps[_statImageDisplay.currentCmap]

        if posCmap is None: posCmap = cmap
        if negCmap is None: negCmap = cmap

        _statImageDisplay.currentCmap += 1
        _statImageDisplay.currentCmap %= len(_statImageDisplay.cmaps)
        opts.cmap                      = posCmap
        opts.negativeCmap              = negCmap

    # The order of these tests is
    # important, due to name overlap

    # P-value image ?
    if any([token in basename for token in pTokens]):
        opts.displayRange  = [0.95, 1.0]
        opts.clippingRange = [0.95, 1.0]

    # T or Z stat image?
    elif any([token in basename for token in statTokens]) and \
       'rendered' not in basename:

        maxVal               = overlay.dataRange[1]
        opts.useNegativeCmap = True
        opts.clippingRange   = [zthres, maxVal]
        opts.displayRange    = [zthres, min((7.5, maxVal))]

    # F stat image?
    elif any([token in basename for token in fStatTokens]):
        opts.displayRange = [0, 10]


# Colour maps used for statistic images
_statImageDisplay.cmaps = ['red-yellow',
                           'blue-lightblue',
                           'green',
                           'cool',
                           'hot',
                           'blue',
                           'red',
                           'yellow',
                           'pink',
                           'copper']


# Index into the cmaps list, pointing to the
# next colour map to use for statistic images.
_statImageDisplay.currentCmap = 0


def _peImageDisplay(overlay, overlayList, displayCtx):
    """Automatically configure display settings for the given PE/COPE
    :class:`.Image` overlay.
    """
    opts = displayCtx.getOpts(overlay)

    opts.cmap            = 'Red-Yellow'
    opts.negativeCmap    = 'Blue-LightBlue'
    opts.useNegativeCmap = True
    opts.displayRange    = [1.0, 100.0]
    opts.clippingRange   = [1.0, overlay.dataRange[1]]


def _FEATImageDisplay(overlay, overlayList, displayCtx):
    """Automatically configure display settings for the given
    :class:`.FEATImage` overlay.
    """
    pass


def _MelodicImageDisplay(overlay, overlayList, displayCtx):
    """Automatically configure display settings for the given
    :class:`.MelodicImage` overlay.
    """

    opts        = displayCtx.getOpts(overlay)
    dmin, dmax  = np.abs(overlay.dataRange)

    # clip low values by default
    if dmax > 3:
        dmin = 3
        dmax = 10

    opts.cmap              = 'Red-Yellow'
    opts.negativeCmap      = 'Blue-LightBlue'
    opts.useNegativeCmap   = True
    opts.displayRange      = [dmin, dmax]
    opts.clippingRange.xlo =  dmin

    # Add the mean as an underlay
    idx      = overlayList.index(overlay)
    meanFile = overlay.getMeanFile()
    existing = [op.abspath(o.dataSource) for o in overlayList]

    # But only if it's not
    # already in the list
    if meanFile not in existing:

        log.debug('Inserting mean melodic image into '
                  'overlay list: {}'.format(meanFile))

        meanImg = fslimage.Image(meanFile)
        overlayList.insert(idx, meanImg)



#!/usr/bin/env python
#
# vectoropts.py - Defines the VectorOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`VectorOpts`, :class:`LineVectorOpts`, and
:class:`RGBVectorOpts` classes, which contain options for displaying NIFTI
vector images.
"""


import copy

import fsleyes_props   as props
import fsl.data.image  as fslimage
import fsleyes.gl      as fslgl
from . import             niftiopts
from . import             volumeopts


class VectorOpts(niftiopts.NiftiOpts):
    """The ``VectorOpts`` class is the base class for :class:`LineVectorOpts`,
    :class:`RGBVectorOpts`, :class:`.TensorOpts`, and :class:`.SHOpts`. It
    contains display settings which are common to each of them.


    *A note on orientation*


    The :attr:`orientFlip` property allows you to flip the left-right
    orientation of line vectors, tensors, and SH functions. This option is
    necessary, because different tools may output vector data in different
    ways, depending on the image orientation.


    For images which are stored radiologically (with the X axis increasaing
    from right to left), the FSL tools (e.g. `dtifit`) will generate vectors
    which are oriented according to the voxel coordinate system. However, for
    neurologically stored images (X axis increasing from left to right), FSL
    tools generate vectors which are radiologically oriented, and thus are
    inverted with respect to the X axis in the voxel coordinate system.
    Therefore, in order to correctly display vectors from such an image, we
    must flip each vector about the X axis.


    This issue is also applicable to ``tensor`` and ``sh`` overlays.
    """


    xColour = props.Colour(default=(1.0, 0.0, 0.0))
    """Colour used to represent the X vector magnitude."""


    yColour = props.Colour(default=(0.0, 1.0, 0.0))
    """Colour used to represent the Y vector magnitude."""


    zColour = props.Colour(default=(0.0, 0.0, 1.0))
    """Colour used to represent the Z vector magnitude."""


    suppressX = props.Boolean(default=False)
    """Do not use the X vector magnitude to colour vectors."""


    suppressY = props.Boolean(default=False)
    """Do not use the Y vector magnitude to colour vectors."""


    suppressZ = props.Boolean(default=False)
    """Do not use the Z vector magnitude to colour vectors."""


    suppressMode = props.Choice(('white', 'black', 'transparent'))
    """How vector direction colours should be suppressed. """


    orientFlip = props.Boolean(default=True)
    """If ``True``, individual vectors are flipped along the x-axis. This
    property is only applicable to the :class:`.LineVectorOpts`,
    :class:`.TensorOpts`, and :class:`.SHOpts` classes. See the
    :meth:`.NiftiOpts.getTransform` method for more information.

    This value defaults to ``True`` for images which have a neurological
    storage order, and ``False`` for radiological images.
    """


    cmap = props.ColourMap()
    """If an image is selected as the :attr:`colourImage`, this colour map
    is used to colour the vector voxels.
    """


    colourImage = props.Choice()
    """Colour vector voxels by the values contained in this image. Any image which
    is in the :class:`.OverlayList`, and which has the same voxel dimensions
    as the vector image can be selected for colouring. If a ``colourImage``
    is selected, the :attr:`xColour`, :attr:`yColour`, :attr:`zColour`,
    :attr:`suppressX`, :attr:`suppressY`, and :attr:`suppressZ` properties are
    all ignored.
    """


    modulateImage  = props.Choice()
    """Modulate the vector colour brightness/alpha by another image. Any image
    which is in the :class:`.OverlayList`, and which has the same voxel
    dimensions as the vector image can be selected for modulation.
    """


    modulateMode = props.Choice(('brightness', 'alpha'))
    """Modulate either the brightness or transparency bu the modulation image.
    """


    clipImage = props.Choice()
    """Clip voxels from the vector image according to another image. Any image
    which is in the :class:`.OverlayList`, and which has the same voxel
    dimensions as the vector image can be selected for clipping. The
    :attr:`clippingRange` dictates the value below which vector voxels are
    clipped.
    """


    clippingRange = props.Bounds(ndims=1)
    """Hide voxels for which the :attr:`clipImage` value is outside of this
    range.
    """


    modulateRange = props.Bounds(ndims=1, clamped=False)
    """Data range used in brightness/transparency modulation, when a
    :attr:`modulateImage` is in use.
    """


    def __init__(self, image, *args, **kwargs):
        """Create a ``VectorOpts`` instance for the given image.  All
        arguments are passed through to the :class:`.NiftiOpts`
        constructor.
        """

        # The orientFlip property defaults to True
        # for neurologically stored images. We
        # give it this vale before calling __init__,
        # because  if this VectorOptse instance has
        # a parent, we want to inherit the parent's
        # value.
        self.orientFlip = image.isNeurological()

        niftiopts.NiftiOpts.__init__(self, image, *args, **kwargs)

        self.__registered = self.getParent() is not None

        if self.__registered:

            self.overlayList.addListener('overlays',
                                         self.name,
                                         self.__overlayListChanged)
            self            .addListener('clipImage',
                                         self.name,
                                         self.__clipImageChanged)
            self            .addListener('modulateImage',
                                         self.name,
                                         self.__modulateImageChanged)

            if not self.isSyncedToParent('modulateImage'):
                self.__refreshAuxImage('modulateImage')
            if not self.isSyncedToParent('clipImage'):
                self.__refreshAuxImage('clipImage')
            if not self.isSyncedToParent('colourImage'):
                self.__refreshAuxImage('colourImage')

        else:
            self.__overlayListChanged()
            self.__clipImageChanged()
            self.__modulateImageChanged()


    def destroy(self):
        """Removes some property listeners, and calls the
        :meth:`.NiftiOpts.destroy` method.
        """
        if self.__registered:
            self.overlayList.removeListener('overlays',      self.name)
            self            .removeListener('clipImage',     self.name)
            self            .removeListener('modulateImage', self.name)

        niftiopts.NiftiOpts.destroy(self)


    def __clipImageChanged(self, *a):
        """Called when the :attr:`clipImage` property changes. Updates
        the range of the :attr:`clippingRange` property.
        """

        image = self.clipImage

        if image is None:
            self.clippingRange.xmin = 0
            self.clippingRange.xmax = 1
            self.clippingRange.x    = [0, 1]
            return

        minval, maxval = image.dataRange

        # Clipping works with <= and >=, so
        # we add an offset allowing the user
        # to configure the overlay such that
        # no voxels are clipped.
        distance = (maxval - minval) / 100.0

        self.clippingRange.xmin =  minval - distance
        self.clippingRange.xmax =  maxval + distance
        self.clippingRange.x    = [minval, maxval + distance]


    def __modulateImageChanged(self, *a):
        """Called when the :attr:`modulateImage` property changes. Updates
        the range of the :attr:`modulateRange` property.
        """

        image = self.modulateImage

        if image is None: minval, maxval = 0, 1
        else:             minval, maxval = image.dataRange

        self.modulateRange.xmin = minval
        self.modulateRange.xmax = maxval
        self.modulateRange.x    = [minval, maxval]


    def __overlayListChanged(self, *a):
        """Called when the overlay list changes. Updates the :attr:`modulateImage`,
        :attr:`colourImage` and :attr:`clipImage` properties so that they
        contain a list of overlays which could be used to modulate the vector
        image.
        """

        overlays = self.displayCtx.getOrderedOverlays()

        # the image for this VectorOpts
        # instance has been removed
        if self.overlay not in overlays:
            return

        self.__refreshAuxImage('modulateImage')
        self.__refreshAuxImage('clipImage')
        self.__refreshAuxImage('colourImage')


    def __refreshAuxImage(self, imageName):
        """Updates the named image property (:attr:`modulateImage`,
        :attr:`colourImage` or :attr:`clipImage`) so that it contains a list
        of overlays which could be used to modulate the vector image.
        """

        prop     = self.getProp(imageName)
        val      = getattr(self, imageName)
        overlays = self.displayCtx.getOrderedOverlays()

        options = [None]

        for overlay in overlays:

            # It doesn't make sense to
            # modulate/clip/colour the
            # image by itself.
            if overlay is self.overlay:
                continue

            # The modulate/clip/colour
            # images must be images.
            if not isinstance(overlay, fslimage.Image):
                continue

            options.append(overlay)

        prop.setChoices(options, instance=self)

        if val in options: setattr(self, imageName, val)
        else:              setattr(self, imageName, None)


class LineVectorOpts(VectorOpts):
    """The ``LineVectorOpts`` class contains settings for displaying vector
    images, using a line to represent the vector value at each voxel.
    """


    lineWidth = props.Real(minval=0.1, maxval=10, default=1, clamped=True)
    """Width of the line in pixels.
    """

    directed = props.Boolean(default=False)
    """If ``True``, the vector data is interpreted as directed. Otherwise,
    the vector data is assumed to be undirected.
    """


    unitLength = props.Boolean(default=True)
    """If ``True``, each vector is scaled so that it has a length of
    ``1 * lengthScale`` (or 0.5 if ``directed`` is ``True``).
    """


    lengthScale = props.Percentage(minval=10, maxval=500, default=100)
    """Length scaling factor. """


    def __init__(self, *args, **kwargs):
        """Create a ``LineVectorOpts`` instance. All arguments are passed
        through  to the :class:`VectorOpts` constructor.
        """

        kwargs['nounbind'] = ['directed', 'unitLength', 'lengthScale']

        VectorOpts.__init__(self, *args, **kwargs)


class RGBVectorOpts(VectorOpts):
    """The ``RGBVectorOpts`` class contains settings for displaying vector
    images, using a combination of three colours to represent the vector value
    at each voxel.
    """


    interpolation = copy.copy(volumeopts.VolumeOpts.interpolation)
    """Apply interpolation to the image data. """


    unitLength = props.Boolean(default=False)
    """If ``True``, the vector data is scaled so it has length 1. """


    def __init__(self, *args, **kwargs):
        """Create a ``RGBVectorOpts`` instance. All arguments are passed
        through  to the :class:`VectorOpts` constructor.
        """

        # We need GL >= 2.1 for
        # spline interpolation
        if float(fslgl.GL_COMPATIBILITY) < 2.1:
            interp = self.getProp('interpolation')
            interp.removeChoice('spline', instance=self)
            interp.updateChoice('linear', instance=self, newAlt=['spline'])

        kwargs['nounbind'] = ['interpolation']
        VectorOpts.__init__(self, *args, **kwargs)



#!/usr/bin/env python
#
# browsexnat.py - Action which allows the user to connect to and browse an
# XNAT repository.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`BrowseXNATAction`, which allows the user
to connect to and browse an XNAT repository. If ``wxnatpy``
(https://github.com/pauldmccarthy/wxnatpy) is not present, the action is
disabled.
"""


import            os
import os.path as op

import wx

import fsl.utils.settings  as fslsettings
import fsleyes.strings     as strings
import fsleyes.autodisplay as autodisplay

from . import base
from . import loadoverlay

# if wxnatpy is not present, the
# action is permanently disabled
try:
    import wxnat
except ImportError:
    wxnat = None


class BrowseXNATAction(base.Action):
    """The ``BrowseXNATAction`` allows the user to open files from an XNAT
    repository. It opens a :class:`XNATBrowser``, and adds the files that
    the user selected into the :class:`.OverlayList`.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``BrowseXNATAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__openBrowser)
        self.__frame = frame

        if wxnat is None:
            self.enabled = False


    def __openBrowser(self):
        """Opens a :class:`XNATBrowser`, then adds any files that the user
        selected to the :class:`.OverlayList`.
        """

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)
            self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[-1]

            if self.displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.overlayList,
                                            self.displayCtx)

        def load(paths):
            if len(paths) == []:
                return
            loadoverlay.loadOverlays(paths,
                                     onLoad=onLoad,
                                     saveDir=False,
                                     inmem=self.displayCtx.loadInMemory)


        dlg = XNATBrowser(self.__frame, load)
        dlg.Layout()
        dlg.Fit()
        dlg.SetSize((-1, 400))
        dlg.CentreOnParent()
        dlg.Show()


class XNATBrowser(wx.Dialog):
    """The ``XNATBrowser`` contains a ``wxnat.XNATBrowserPanel``, allowing the
    user to connect to and browse an XNAT repository. It contains a *Download*
    button which, when clicked, downloads all selected files from the
    repository into a temporary directory, and passes the file paths to a
    provided callback function.
    """


    def __init__(self, parent, loadFunc=None):
        """Create a ``XNATBrowser``.

        :arg parent:   ``wx`` parent object

        :arg loadFunc: Function to call when the user has downloaded
                       some files. Passed a list of files paths.
        """

        if wxnat is None:
            raise RuntimeError('wxnatpy is not available!')

        wx.Dialog.__init__(self,
                           parent,
                           title=strings.titles[self],
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        hosts    = fslsettings.read('fsleyes.xnat.hosts',    [])
        accounts = fslsettings.read('fsleyes.xnat.accounts', {})
        filters  = {'file' : '*.nii|*.nii.gz|*.img|*.img.gz|*.gii|*.vtk'}

        self.__loadFunc = loadFunc
        self.__destDir  = None
        self.__panel    = wxnat.XNATBrowserPanel(
            self,
            knownHosts=hosts,
            knownAccounts=accounts,
            filterType='glob',
            filters=filters)

        self.__download     = wx.Button(self, wx.ID_OK)
        self.__close        = wx.Button(self, wx.ID_CANCEL)
        self.__fullPathsLbl = wx.StaticText(self)
        self.__fullPaths    = wx.CheckBox(self)
        self.__ctrlSizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer        = wx.BoxSizer(wx.VERTICAL)

        self.__fullPathsLbl.SetLabel(strings.labels[self, 'fullPaths'])
        self.__download    .SetLabel(strings.labels[self, 'download'])
        self.__close       .SetLabel(strings.labels[self, 'close'])

        self.__ctrlSizer.Add((10, 1),             flag=wx.EXPAND, proportion=1)
        self.__ctrlSizer.Add(self.__fullPathsLbl, flag=wx.EXPAND)
        self.__ctrlSizer.Add(self.__fullPaths,    flag=wx.EXPAND)
        self.__ctrlSizer.Add((10, 1),             flag=wx.EXPAND)
        self.__ctrlSizer.Add(self.__download,     flag=wx.EXPAND)
        self.__ctrlSizer.Add((10, 1),             flag=wx.EXPAND)
        self.__ctrlSizer.Add(self.__close,        flag=wx.EXPAND)
        self.__ctrlSizer.Add((10, 1),             flag=wx.EXPAND)

        self.__sizer.Add((1, 10),          flag=wx.EXPAND)
        self.__sizer.Add(self.__panel,     flag=wx.EXPAND, proportion=1)
        self.__sizer.Add((1, 10),          flag=wx.EXPAND)
        self.__sizer.Add(self.__ctrlSizer, flag=wx.EXPAND)
        self.__sizer.Add((1, 10),          flag=wx.EXPAND)

        self.SetSizer(self.__sizer)
        self.__download.SetDefault()

        self.__panel   .Bind(wxnat.EVT_XNAT_ITEM_HIGHLIGHT_EVENT,
                             self.__onHighlight)
        self.__panel   .Bind(wxnat.EVT_XNAT_FILE_SELECT_EVENT,
                             self.__onDownload)
        self.__download.Bind(wx.EVT_BUTTON, self.__onDownload)
        self.__close   .Bind(wx.EVT_BUTTON, self.__onCloseButton)
        self           .Bind(wx.EVT_CLOSE,  self.__onClose)

        self.__fullPaths.SetValue(True)

        # ok/download button only enabled when file
        # items are selected in the tree view
        self.__download.Disable()


    def GetHosts(self):
        """Wraps ``wxnat.XNATBrowserPanel.GetHosts``. """
        return self.__panel.GetHosts()


    def GetAccounts(self):
        """Wraps ``wxnat.XNATBrowserPanel.GetAccounts``. """
        return self.__panel.GetAccounts()


    def __onHighlight(self, ev):
        """Called when the item selection in the tree browser is changed.
        Enables/disables the download button depending on whether any
        files are highlighted.
        """
        self.__download.Enable(len(self.__panel.GetSelectedFiles()) > 0)


    def __onDownload(self, ev):
        """Called when the *Download* button is pushed. Prompts the user to
        select a directory, and then downloads the files.
        """

        files = self.__panel.GetSelectedFiles()

        if len(files) == 0:
            return

        destDir = self.__destDir

        if destDir is None:
            destDir = fslsettings.read('fsleyes.xnat.downloaddir', os.getcwd())
            dlg     = wx.DirDialog(self,
                                   strings.labels[self, 'choosedir'],
                                   defaultPath=destDir)

            if dlg.ShowModal() != wx.ID_OK:
                return

            destDir        = dlg.GetPath()
            self.__destDir = destDir

        paths = []

        for f in files:
            if self.__fullPaths.GetValue():
                dest = wxnat.generateFilePath(f)
                dest = op.join(destDir, dest)
                os.makedirs(op.dirname(dest), exist_ok=True)
            else:
                dest = op.join(destDir, op.basename(dest))

            dest = self.__panel.DownloadFile(f, dest)

            if dest is not None:
                paths.append(dest)

        if self.__loadFunc is not None:
            self.__loadFunc(paths)

        fslsettings.write('fsleyes.xnat.downloaddir', destDir)


    def __onClose(self, ev):
        """Called on EVT_CLOSE events. Destroys this dialog. """
        self.__panel.EndSession()
        fslsettings.write('fsleyes.xnat.hosts',    self.GetHosts())
        fslsettings.write('fsleyes.xnat.accounts', self.GetAccounts())

        if self.IsModal(): self.EndModal(wx.ID_CANCEL)
        else:              self.Destroy()


    def __onCloseButton(self, ev):
        """Called when the *Close* button is pushed. Closes the dialog. """
        self.Close()



#!/usr/bin/env python
#
# savelayout.py - The SaveLayoutAction
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SaveLayoutAction` class, an action
which allows the user to save the current layout.
"""


import fsleyes.strings as strings
import fsleyes.layouts as layouts
from . import             base


class SaveLayoutAction(base.Action):
    """The ``SaveLayoutAction`` allows the user to save the current
    :class:`.FSLeyesFrame` layout, so it can be restored
    at a later time. See the :mod:`.layouts` module.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``SaveLayoutAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        self.__frame = frame
        base.Action.__init__(self, overlayList, displayCtx, self.__saveLayout)


    def __saveLayout(self):
        """Save the current :class:`.FSLeyesFrame` layout.  The user is
        prompted to enter a name, and the current frame layout is saved
        via the :func:`.layouts.saveLayout` function.
        """

        import wx

        builtIns = list(layouts.BUILT_IN_LAYOUTS.keys())
        saved    = layouts.getAllLayouts()

        while True:
            dlg = wx.TextEntryDialog(
                self.__frame,
                message=strings.messages[self, 'enterName'])

            if dlg.ShowModal() != wx.ID_OK:
                return

            name = dlg.GetValue()

            if name.strip() == '':
                return

            # Not allowed to use built-in layout names
            if name in builtIns:
                dlg = wx.MessageDialog(
                    self.__frame,
                    message=strings.messages[
                        self, 'nameIsBuiltIn'].format(name),
                    style=(wx.ICON_EXCLAMATION | wx.OK))
                dlg.ShowModal()
                continue

            # Name collision - confirm overwrite
            if name in saved:
                dlg = wx.MessageDialog(
                    self.__frame,
                    message=strings.messages[
                        self, 'confirmOverwrite'].format(name),
                    style=(wx.ICON_QUESTION | wx.YES_NO | wx.NO_DEFAULT))

                if dlg.ShowModal() == wx.ID_NO:
                    continue

            break

        layouts.saveLayout(self.__frame, name)

        self.__frame.refreshLayoutMenu()



#!/usr/bin/env python
#
# cliserver.py - Infrastructure to call an existing FSLeyes instance from the
#                command line.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module implements a simple client-server architecture which can be
used to call an existing FSLeyes instance from the command line.


When FSLeyes is started with the ``--cliserver`` option, the :func:`runserver`
function is called to start a server thread. On subsequent invocations of
FSLeyes (with the ``--cliserver`` option), instead of starting up a new
FSLeyes instance, the command-line arguments are passed to the original
instance via the :func:`send` function.


When the :func:`runserver` function is called, it uses the
:mod:`fsl.utils.settings` module to save the TCP port number to a file called
:func:`cliserver.txt`. This file is used by the :func:`send` function to
determine the port to connect to, and by the :func:`isRunning` function to
determine whether or not a server is running.
"""


import os
import sys
import shlex
import atexit
import socket
import logging
import argparse
import threading

import fsl.utils.settings               as fslsettings
import fsleyes.actions.applycommandline as applycli


log = logging.getLogger(__name__)


class CLIServerAction(argparse.Action):
    """Custom ``argparse.Action`` for applying the ``--cliserver`` command-line
    option.

    If a server is not running, the ``namespace.cliserver`` attribute is set
    to ``True``. Otherwise, the remaining arguments are passed to the
    :func:`send` function, and the process is closed via ``sys.exit``.

    In the former case (a server is not already running), the
    :mod:`fsleyes.main` module will start a server via :func:`runserver` at
    a later point in time.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``CLIServerAction``.

        :arg allArgs: Sequence of arguments that should be passed to the
                      :func:`send` function, if this action is invoked as a
                      client. If ``None``, it is set to ``sys.argv[1:]``.

        All other arguments are passed to ``argparse.Action.__init__``.
        """

        kwargs['nargs'] = 0
        allArgs = kwargs.pop('allArgs', None)

        if allArgs is None:
            allArgs = sys.argv[1:]

        self.__allArgs = allArgs

        argparse.Action.__init__(self, *args, **kwargs)


    def __call__(self, parser, namespace, values, option_string):

        if not isRunning():
            setattr(namespace, self.dest, True)
            return

        else:
            # first arg is the current working
            # directory - see runserver
            argv = list(self.__allArgs)
            argv.remove(option_string)
            argv = [os.getcwd()] + argv
            line = ' '.join(shlex.quote(a) for a in argv)
            send(line)
            sys.exit(0)


class AlreadyRunningError(Exception):
    """Raised by :func:`runserver` if a server loop is already running. """
    pass


class NotRunningError(Exception):
    """Raised by :func:`send` if a server loop is not running. """
    pass


def runserver(overlayList, displayCtx, ev=None):
    """Starts a thread which runs the :func:`_serverloop` function.

    If a server is already running, within this or any other FSLeyes instance,
    an :exc:`AlreadyRunningError` is raised.

    Every line that is received is assumed to contain command line
    arguments specifying overlays to be loaded; these are passed
    to the :func:`.applyCommandLineArgs` function.

    :arg overlayList: The :class:`OverlayList`
    :arg displayCtx:  The master :class:`DisplayContext`
    :arg ev:          Optional ``threading.Event`` which can be used
                      to terminate the server thread.
    """

    if isRunning():
        raise AlreadyRunningError()

    def callback(line):

        # first arg is the directory that
        # the client was executed from,
        # which is used by applyCLIArgs
        # in case overlays were specified
        # with relative paths
        args    = shlex.split(line)
        baseDir = args[0]
        args    = args[1:]

        applycli.applyCommandLineArgs(overlayList,
                                      displayCtx,
                                      args,
                                      baseDir=baseDir)

    t        = threading.Thread(target=_serverloop, args=(callback, ev))
    t.daemon = True

    t.start()


def isRunning():
    """Returns ``True`` if (it looks like) a server is running, ``False``
    otherwise.
    """
    return fslsettings.readFile('cliserver.txt') is not None


def _serverloop(callback, ev=None):
    """Starts a TCP server which runs forever.

    The server port number is written to the FSLeyes settings directoy in a
    file called ``cliserver.txt`` (see :mod:`fsl.utils.settings`).  Then,
    every line of text received on the socket is passed to the ``callback``
    function.

    :arg callback: Callback function to which every line that is received
                   is passed.
    :arg ev:       Optional ``threading.Event`` which can be used to signal
                   the server thread to stop.
    """

    if ev is None:
        ev = threading.Event()

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('localhost', 0))
    sock.listen(1)
    sock.settimeout(1)
    port = sock.getsockname()[1]

    with fslsettings.writeFile('cliserver.txt') as f:
        f.write('{}'.format(port))

    atexit.register(fslsettings.deleteFile, 'cliserver.txt')

    log.debug('CLI server running on port %i', port)

    while True:

        try:
            conn, addr = sock.accept()
        except socket.timeout:
            if ev.isSet(): break
            else:          continue

        log.debug('Connection from %s', addr)

        with conn:

            line = conn.makefile().readline().strip()

            log.debug('Received %s ...', line[:50])

            try:
                callback(line)
            except Exception as e:
                log.warning('Callback function raised error: %s',
                            e, exc_info=True)


def send(line):
    """If a cli server is running (see :func:`runserver` and
    :func:`_serverloop`), the given ``args`` are sent to it.

    A :exc:`NotRunningError` is raised if a server loop is not running.
    """

    if not isRunning():
        raise NotRunningError()

    with fslsettings.use(fslsettings.Settings('fsleyes', writeOnExit=False)):
        port = int(fslsettings.readFile('cliserver.txt').strip())

    line = (line + '\n').encode()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(('localhost', port))

    log.debug('Sending to port %i: %s...', port, line[:50])

    sock.sendall(line)
